

#ifndef COMPONENT_H
#define COMPONENT_H


#include "structure.h"

#ifdef SYMBOLIC
#include "cudd.h"
#endif

class Component;
  class Atom;
  class Compound;

class Port;
class Connector;
class Clock;
class GenericClock;
class LogicalClock;
class Constraint;
class Priority;
class Scheduler;
class Node;
class CtrlState;

/***********************************
 * Class: Component
 * Description: component base class
 ***********************************/
class Component {

  /*** Protected data members ***/
  protected:
    char*         mName;   /* Name of the component */
    Compound*     mHolder; /* Holder component, if any */
    GenericClock* mClock;
	
  public:
    //Array<Port,false> Envports ;

  /*** Constructors & Dectructors ***/
  public:
    Component(const char*, 
              const Compound*);
    virtual ~Component();


  /*** Protected Data-Access Methods ***/
  public:
    inline void          setClock(GenericClock* clock) { mClock = clock; }

  /*** Public Data-Access Methods ***/
  public:    
    inline char*         name()   const { return mName; }
    inline Compound*     holder() const { return mHolder; }	
    inline GenericClock* clock()  const { return mClock; }

  /*** Public Utility Methods ***/
  public:
    virtual bool       isAtom()     const { return false; }
    virtual bool       isCompound() const { return false; }
    virtual void       initialize()                    = 0;
    virtual void       initialize(LogicalClock *clock) = 0;

  /*** Public Methods invoked from appln. ***/
    virtual Node*      getState()      = 0;
    virtual void       setState(Node*) = 0;
};

/***********************************
 * Class: Atom
 * Description: Atomic component 
 ***********************************/
class Atom : public Component {

  /*** Protected data members ***/
  protected:
    Array<Port,true>       mPorts;       /* Atom ports */
    Array<Clock,true>      mClocks;      /* Atom clocks */
    Array<Constraint,true> mConstraints; /* Atom time constraints */
    Array<Connector,false> mLinks;       /* Atom links (connectors) */
    Array<CtrlState, true> mCtrlStates;/* control-states */
    CtrlState*             BIP_STATE;  /* current control-state */
    Scheduler*             mScheduler; /* back pointer to scheduler */
#ifdef SYMBOLIC
    DdNode*                mBdd;       /* generic atomic behavior BDD*/
    DdNode*                mStateBdd;  /* current state BDD */
#endif

  /*** Constructoris & Dectructors ***/
  public:
    Atom(const char*, 
         const Compound*);
    virtual ~Atom();
 
  /*** Public Data-Access Methods ***/
  public:
    inline Array<Port,true>&       ports()       { return mPorts; }
    inline Array<Clock,true>&      clocks()      { return mClocks; }
    inline Array<Constraint,true>& constraints() { return mConstraints; }
    inline Array<Connector,false>& links()       { return mLinks; }
    inline Array<CtrlState, true>& ctrlStates()  { return mCtrlStates; }

  /*** Public Utility Methods ***/
  public:
    virtual bool       isAtom() const     { return true; }
            void       notifyDependentConn();
            void       resetLink();      /* mark links for re-evaluation */
            void       addPort(Port *p); /*      */
#ifdef SYMBOLIC
    inline  DdNode*    bdd()              { return mBdd; } 
    inline  DdNode*    stateBdd()         { return mStateBdd; } 
            void       initBdd();         /* initialize port-bdd's of this atom */
            void       initStateBdd();    /* initialize control-state bdds */
            DdNode*    createAtomicBdd(); /* routine called by initBehavBdd() */
            void       initBehavBdd();    /* creates behavior-bdd of this atom */
#endif

  /*** Public Methods visible from appln. ***/
  public:
    virtual void  initialize()                    { }               /* for compatibility with the older code generator */
    virtual void  initialize(LogicalClock *clock) { initialize(); } /* for compatibility with the older code generator */
    virtual void  execute(Port* port)  = 0;
    virtual Node* getState()           = 0;
    virtual void  setState(Node*)      = 0;
    virtual Port* sync(int nPorts, ...);
    virtual Port* rt_sync(int nPorts, ...);
    static  void  computePortConstraint(Constraint**);
    static  void  applyPriority(Port*, Constraint*, Port*, Constraint*);
#ifdef SYMBOLIC
            void  updateStateBdd(DdNode* bdd);
            void  resetStateBdd();
#endif

};

/***********************************
 * Class: Compound
 * Description: Compound component
 ***********************************/
class Compound : public Component {

  /*** Protected data members ***/
  protected:
    Array<Component,true> mComponents; /* subcomponents */
    Array<Connector,true> mConnectors; /* local connectors */
    Array<Priority,true>  mPriorities; /* local priorities */

    Connector*            mLegals;     /* list of synced legal connectors */
    Port*                 mInternals;  /* list of synced internal ports   */

  /*** Constructors & Dectructors ***/
  public:
  Compound(const char*, 
           const Compound*);
  virtual ~Compound();

  /*** Public Data-Access Methods ***/
  public:
    inline Array<Connector,true>& connectors()            { return mConnectors; }
    inline Array<Priority,true>&  priorities()            { return mPriorities; }
    inline Port*                  internals()       const { return mInternals; }
    inline void                   setInternals(Port* p)   { mInternals = p; }

  /*** Public Utility Methods ***/
  public:
    virtual bool       isCompound() const { return true; }
    virtual void       initialize();
    virtual void       initialize(LogicalClock *);
            Connector* subactivate(bool);
            Connector* activate(bool);
            void       initPrioHeap(Scheduler*);
            void       setup(const Scheduler*);
            void       setStaticPrioLevel();
#ifdef SYMBOLIC
            void       initBdd();
            void       initAtomicBdds();
            void       initConnectorBdds();
            void       initAtomStateBdds();
            DdNode*    createConnBdd();
            DdNode*    createBehavBdd();
            DdNode*    createAtomicStateBdd();
#endif

  /*** Public Methods invoked from appln. ***/
  public:
    virtual Node*     getState();
    virtual void      setState(Node*);
    static  Compound* Create(int, char**);
};


class CtrlState {

  /*** Protected data members ***/
  protected:
    char*   mName;
#ifdef SYMBOLIC
    DdNode* mBdd;
#endif
  /*** Constructoris & Dectructors ***/
  public:
    CtrlState(const char* = NULL);
    ~CtrlState();

  /*** Public Data-Access Methods ***/
#ifdef SYMBOLIC
    inline  DdNode*    bdd()   { return mBdd; }
#endif
  /*** Public Utility Methods ***/
#ifdef SYMBOLIC
    void    initBdd();
#endif
};

#endif

