


#ifndef CONNECTOR_H
#define CONNECTOR_H

#include "structure.h"
#include "constraint.h"
#include "executable.h"

#include <stdarg.h>

#ifdef SYMBOLIC
  #include "cudd.h"
#endif

class Interaction;
class Compound;
class Port;
class Priority;

/*******************************
 * Class: Connector
 * Description: Connector class
 *******************************/
class Connector {
  /*** Private data members ***/
    static char LABEL[1024];

  /*** Protected data members ***/
  protected:
    char*              mName;           /* name of connector */
    Compound*          mHolder;         /* holder compound */
	Array<Port, false> mPorts;         /* port array */
    Port*              mxPort;          /* exported port */
    bool               mLegal;          /* legal i.e. complete or 
                                          maximal */
    bool               mDominated;      /* dominated connector (no rt-constraint, a priority disable the (only) enabled interaction) */
    Array<Priority, false> mDominatedPriorities;  /* (jcombaz) */
    Array<Priority, false> mDominatorPriorities;  /* (jcombaz) */

    Interaction*       mFeasibleInter;  /* feasible interaction */
    bool               mSubconn;        /* used as a sub-connector
                                          in hierarcical connector */
    bool               mSubconnSameHolder; /* used as a sub-connector
					     in hierarcical connector
                                             which is has the same holder */
    int                mPriority;       /* static priority level */
    bool               mQueued;         /* enqueued in the priority 
                                          heap */
    Connector*         mNext;           /* next legal */
    bool               mEval;           /* flag marking evaluated */
    bool               mRecomputexPort; /* recomputation of xPort() needed (jcombaz) */
    Array<Connector, false> mLinks;     /* back pointer to higher links */
#ifdef SYMBOLIC
    DdNode*            mBdd;            /* BDD for Connector */
#endif

  /*** Constructors & Dectructors ***/
  public:
    Connector(const char*, 
              const Compound*, 
              const bool, 
              const int, ...);
    Connector(const char*,
              const Compound*,
              const bool);
    virtual ~Connector();

  /*** Public Data-Access Methods ***/
  public:
    inline char*               name()      const { return mName; }
    inline Compound*           holder()    const { return mHolder; }
    inline Array<Port, false>& ports()           { return mPorts; }
    
    inline Port*               xPort()     const { return mxPort; }
    inline void                setxPort(Port* p) { mxPort = p; }
    inline bool                recomputexPort() const { return mRecomputexPort; }        /* (added by jcombaz) */
    inline void                setRecomputexPort(const bool b) { mRecomputexPort = b; }  /* (added by jcombaz) */
    inline bool                legal()     const { return mLegal; }
    inline bool                dominated() const { return mDominated; }
    inline void                setDominated(bool b) { mDominated = b; }
    inline Array<Priority, false>& dominatedPriorities() { return mDominatedPriorities; } /* (added by jcombaz) */
    inline Array<Priority, false>& dominatorPriorities() { return mDominatorPriorities; } /* (added by jcombaz) */

    inline Interaction*        feasibleInter() const
                                 { return mFeasibleInter; }                        
    inline void                setFeasibleInter(Interaction* i)
                                 { mFeasibleInter = i; }
    inline bool                isSubconn() const   { return mSubconn; }
    inline bool                isSubconnSameHolder() const { return mSubconnSameHolder; }
    inline void                setSubconnSameHolder(bool b) { mSubconnSameHolder = b; }
    inline int                 getPriority() const { return mPriority; }
    inline void                setPriority(int p)  { mPriority = p; }
    inline bool                noPrio()    const   
                                     { return (mPriority == -9999); }
    inline bool                isQueued()  const   { return mQueued; }
    inline void                setQueued(const bool b) { mQueued = b; }
    inline Connector*          next()      const   { return mNext; }
    inline void                setNext(Connector* next) 
                                 { mNext = next; }
    inline bool                evaluated() const   { return mEval; }
    inline void                setEvaluated(bool b){ mEval = b; }
    inline Array<Connector, false>& links()        { return mLinks; }
#ifdef SYMBOLIC
    inline DdNode*             bdd() const         { return mBdd; }
#endif
    virtual bool               isBroadcast() const { return true; }
            

   /*** Public Utility Methods ***/
   public:
     char*        label();
     char*        rt_label(const BitVector&);
     void         eval(bool opt);
     void         computexPort();
     void         setDominatedPriority(Priority*);
     void         setDominatorPriority(Priority*);
     Connector*   evaluate(Connector* last, bool isSubConn, bool opt);
     void         execute();
     void         execute(const BIP::Time&);
     void         actuate(const BIP::Time&);
     //void         notifyDependentPorts();
     void         createFeasibleInter();  /* create feasible interactions at the current state */
     void         createFeasibleInters(); /* create feasible interactions depending of the real-time (added by jcombaz) */
     Interaction* createFeasibleBroadcast(BitVector&,BitVector&,int,Interaction*);
     void         resetLink();           // reset this and dependents
     void         resetDependentLinks(); // reset dependents
     virtual bool complete() const;
     virtual bool guard();
     virtual bool guard(const BitVector&);
     bool         maximal() const;
     virtual void up();
     virtual void up(const BitVector&);
     virtual void dn();
     virtual void dn(const BitVector&);

     void         setVarPorts(const int, va_list);
#ifdef SYMBOLIC
     void         initBdd();
     void         defineBdd(int, ...);
#endif

   /*** Public Methods invoked from appln. ***/
   public:
     virtual void define() {}
};


/*********************************
 * Class: Interaction
 * Description: Interaction class
 *********************************/
class Interaction: public Executable {

  /*** Protected data members ***/
  protected:
    Connector*            mConnector;  /* the parent Connector */
    BitVector             mSynced;     /* the synced ports     */
    Array<Priority,false> mInhibitors; /* potential inhibitors */

    Interaction*          mNext;        /* next legal interaction (added by jcombaz) */

    static HashTable<Interaction> TABLE; 

  /*** Constructors & Dectructors ***/
  public:
    Interaction(const Connector*, 
                const BitVector&);
    Interaction(const Interaction&);
    virtual ~Interaction();

  /*** Public Data-Access Methods ***/
  public:
    inline Connector*             connector() const 
                                    { return mConnector; }
    inline const BitVector&       synced()    const 
                                    { return mSynced; }
    inline Array<Priority,false>& inhibitors()      
                                    { return mInhibitors; }
    inline Interaction* next()                                const { return mNext; }
    inline void         setNext(Interaction* interaction)           { mNext = interaction; }

  /*** Public Utility Methods ***/
  public:
    bool         equal(const Interaction*)    const;
    bool         isSubset(const Interaction*, 
                          bool&)              const;
    void         execute(const BIP::Time&);
    void         actuate(const BIP::Time&);
    BIP::Time    deadline(const BIP::Time&)   const;
    BIP::Time    next(const BIP::Time&)       const;
    GenericClock* displayClock()              const;
    //void         notifyDependentPorts();
    Interaction* copy()                       const;
    unsigned     hash(const unsigned)         const;

    void         print()                      const;
   
    static Interaction* Create(const Connector*,  
                               const int, ...);
    static Interaction* Create(const Connector*, 
                               const BitVector&);
    static Interaction* CreateRealTime(const Connector*,const BitVector&,const BitVector&);

    char*        label();
};

#endif
